@@ -34,6 +34,7 @@ best_docs = Document.match('some phrase')
 import inspect
 import math
 import os
+import re
 import struct
 import sys
 try:
@@ -293,7 +294,7 @@ class FTSModel(BaseFTSModel):
     docid = DocIDField()
 
     @classmethod
-    def validate(cls):
+    def validate_model(cls):
         if cls._meta.primary_key.name != 'docid':
             raise ImproperlyConfigured(
                 'FTSModel classes must use the default `docid` primary key.')
@@ -351,7 +352,8 @@ class FTSModel(BaseFTSModel):
         return fn.fts_bm25(match_info, *weights)
 
     @classmethod
-    def _search(cls, term, weights, with_score, score_alias, score_fn):
+    def _search(cls, term, weights, with_score, score_alias, score_fn,
+                explicit_ordering):
         if not weights:
             rank = score_fn()
         elif isinstance(weights, dict):
@@ -367,6 +369,7 @@ class FTSModel(BaseFTSModel):
         order_by = rank
         if with_score:
             selection = (cls, rank.alias(score_alias))
+        if with_score and not explicit_ordering:
             order_by = SQL(score_alias)
 
         return (cls
@@ -375,15 +378,28 @@ class FTSModel(BaseFTSModel):
                 .order_by(order_by))
 
     @classmethod
-    def search(cls, term, weights=None, with_score=False, score_alias='score'):
+    def search(cls, term, weights=None, with_score=False, score_alias='score',
+               explicit_ordering=False):
         """Full-text search using selected `term`."""
-        return cls._search(term, weights, with_score, score_alias, cls.rank)
+        return cls._search(
+            term,
+            weights,
+            with_score,
+            score_alias,
+            cls.rank,
+            explicit_ordering)
 
     @classmethod
     def search_bm25(cls, term, weights=None, with_score=False,
-                    score_alias='score'):
+                    score_alias='score', explicit_ordering=False):
         """Full-text search for selected `term` using BM25 algorithm."""
-        return cls._search(term, weights, with_score, score_alias, cls.bm25)
+        return cls._search(
+            term,
+            weights,
+            with_score,
+            score_alias,
+            cls.bm25,
+            explicit_ordering)
 
 
 class SearchField(BareField):
@@ -399,6 +415,26 @@ class SearchField(BareField):
             unindexed=self._unindexed, **kwargs)
 
 
+_alphabet = 'abcdefghijklmnopqrstuvwxyz'
+_alphanum = set([
+    '\t',
+    chr(26),  # Substitution control character.
+    '"',
+    '(',
+    ')',
+    '*',
+    ':',
+    '_',
+    ' ',
+    '+',
+    ',',
+    '{',
+    '}',
+]) | set('0123456789') | set(_alphabet) | set(_alphabet.upper())
+_invalid_ascii = set([chr(p) for p in range(128) if chr(p) not in _alphanum])
+_quote_re = re.compile('(?:[^\s"]|"(?:\\.|[^"])*")+')
+
+
 class FTS5Model(BaseFTSModel):
     """
     Requires SQLite >= 3.9.0.
@@ -466,7 +502,7 @@ class FTS5Model(BaseFTSModel):
     rowid = RowIDField()
 
     @classmethod
-    def validate(cls):
+    def validate_model(cls):
         # Perform FTS5-specific validation and options post-processing.
         if cls._meta.primary_key.name != 'rowid':
             raise ImproperlyConfigured(cls._error_messages['pk'])
@@ -498,6 +534,40 @@ class FTS5Model(BaseFTSModel):
 
         return True
 
+    @staticmethod
+    def validate_query(query):
+        tokens = _quote_re.findall(query)
+        for token in tokens:
+            if token.startswith('"') and token.endswith('"'):
+                continue
+            if set(token) & _invalid_ascii:
+                return False
+        return True
+
+    @staticmethod
+    def clean_query(query, replace=chr(26)):
+        """
+        Clean a query of invalid tokens.
+        """
+        accum = []
+        any_invalid = False
+        tokens = _quote_re.findall(query)
+        for token in tokens:
+            if token.startswith('"') and token.endswith('"'):
+                accum.append(token)
+                continue
+            token_set = set(token)
+            invalid_for_token = token_set & _invalid_ascii
+            if invalid_for_token:
+                any_invalid = True
+                for c in invalid_for_token:
+                    token = token.replace(c, replace)
+            accum.append(token)
+
+        if any_invalid:
+            return ' '.join(accum)
+        return query
+
     @classmethod
     def match(cls, term):
         """
@@ -517,13 +587,19 @@ class FTS5Model(BaseFTSModel):
         return fn.bm25(cls.as_entity(), *weights)
 
     @classmethod
-    def search(cls, term, weights=None, with_score=False, score_alias='score'):
+    def search(cls, term, weights=None, with_score=False, score_alias='score',
+               explicit_ordering=False):
         """Full-text search using selected `term`."""
-        return cls.search_bm25(term, weights, with_score, score_alias)
+        return cls.search_bm25(
+            FTS5Model.clean_query(term),
+            weights,
+            with_score,
+            score_alias,
+            explicit_ordering)
 
     @classmethod
     def search_bm25(cls, term, weights=None, with_score=False,
-                    score_alias='score'):
+                    score_alias='score', explicit_ordering=False):
         """Full-text search using selected `term`."""
         if not weights:
             rank = SQL('rank')
@@ -540,11 +616,12 @@ class FTS5Model(BaseFTSModel):
         order_by = rank
         if with_score:
             selection = (cls, rank.alias(score_alias))
+        if with_score and not explicit_ordering:
             order_by = SQL(score_alias)
 
         return (cls
                 .select(*selection)
-                .where(cls.match(term))
+                .where(cls.match(FTS5Model.clean_query(term)))
                 .order_by(order_by))
 
     @classmethod
